"""
Entry points for all of the primary SRM commands.
"""
import click

from . import config


@click.command()
def init() -> None:
    """Initialize the current directory for SRM."""
    conf = config.LocalConf()
    if not conf.exists():
        conf.load(create=True)
    else:
        click.secho("Directory is already initialized!", fg='red')


@click.command()
def status() -> None:
    """Show ROM state and info."""
    conf = config.LocalConf()
    if not conf.exists():
        raise click.ClickException("Directory is not initialized! Try the 'init' command.")
    click.echo("temp status")



"""
Test cases for file operations.
"""

import tempfile

from srm.file import Path


def test_path_calculates_hashes():
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(b'The quick brown fox jumps over the lazy dog')
        f.close()
        p = Path(f.name)
        assert p.crc() == '414fa339'
        assert p.md5() == '9e107d9d372bb6826bd81d3542a419d6'
        assert p.sha1() == '2fd4e1c67a2d28fced849ee1bb76e7391b93eb12'
        p.unlink()



import os
import tempfile
from unittest.mock import patch, sentinel

import pytest
import toml

from srm.config import ChainConf, Conf, GlobalConf, LocalConf

# value for testing set/get
V = sentinel.V

def test_config_not_exists():
    with tempfile.TemporaryDirectory() as d:
        c = Conf(d)
        assert not c.exists()


def test_config_exists():
    with tempfile.NamedTemporaryFile() as t:
        c = Conf(t.name)
        assert c.exists()


def test_load_create_with_path():
    with tempfile.TemporaryDirectory() as d:
        c = Conf(os.path.join(d, "a/b/file.txt"))
        c.load(create=True)
        assert c.exists()


def test_set_and_get_key():
    c = Conf('cofig')
    c['key'] = V
    assert c['key'] == V


def test_set_key_not_in_whitelist_raises_exception():
    c = Conf('cofig', valid_keys=['field_ok'])
    with pytest.raises(KeyError):
        c['field_bad'] = V


def test_set_and_del_key():
    c = Conf('config')
    c['key'] = V
    del c['key']
    with pytest.raises(KeyError):
        c['key']


def test_set_and_get_nested_key():
    c = Conf('config')
    c['s1.s2.key'] = V
    assert c['s1.s2.key'] == V


def test_set_and_del_nested_key():
    c = Conf('config')
    c['s1.s2.key'] = V
    del c['s1.s2.key']
    with pytest.raises(KeyError):
        c['s1.s2.key']


def test_dump_config():
    with tempfile.NamedTemporaryFile() as t:
        c = Conf(t.name)
        c['key'] = V
        c.dump()
        with open(t.name) as f:
            assert f.readline() == 'key = sentinel.V\n'


def test_dump_nested_config():
    with tempfile.NamedTemporaryFile() as t:
        c = Conf(t.name)
        c['s1.s2.key'] = V
        assert 's1.s2.key' in c
        c.dump()
        with open(t.name) as f:
            assert f.readlines() == ['[s1.s2]\n', 'key = sentinel.V\n']


def test_load_config():
    with tempfile.NamedTemporaryFile() as t:
        with open(t.name, 'w+') as f:
            toml.dump({'key1': 1, 'key2': 2}, f)
        c = Conf(t.name)
        c.load()
        assert 'key2' in c
        assert c['key2'] == 2


def test_dump_chain_config():
    with tempfile.NamedTemporaryFile() as t:
        c = ChainConf(Conf(t.name))
        c['key'] = V
        c.dump()
        with open(t.name) as f:
            assert f.readline() == 'key = sentinel.V\n'


def test_load_chain_config():
    with tempfile.NamedTemporaryFile() as t:
        with open(t.name, 'w+') as f:
            toml.dump({'key1': 1, 'key2': 2}, f)
        c = ChainConf(Conf(t.name))
        c.load()
        assert 'key2' in c
        assert c['key2'] == 2


def test_get_global_key_from_chain_config():
    with tempfile.NamedTemporaryFile() as global_path:
        # set a key/value into mock global config
        gconf = Conf(global_path.name)
        gconf['key'] = 1
        gconf.dump()
        # test
        with tempfile.NamedTemporaryFile() as local_path:
            c = ChainConf(Conf(local_path.name), gconf)
            c.load()
            assert 'key' in c
            assert c['key'] == 1

def test_chain_config_exists():
    with tempfile.NamedTemporaryFile() as t:
        c = ChainConf(Conf(t.name))
        assert c.exists()



"""
Manage global and local configuration data.
"""

import collections
import os
from typing import Dict  # pylint: disable=unused-import
from typing import Any, Iterator, List, Optional, Set, Tuple, cast

import toml  # type: ignore
from boltons import funcutils, iterutils  # type: ignore

# pylint: disable=notimplemented-raised, raising-bad-type

_LOCAL_PATH = ".srm/config"

_GLOBAL_PATH = "~/.srmconfig"
_GLOBAL_KEYS = {'my.temp.key'}


class Conf(collections.abc.MutableMapping):
    """Config store, backed by a TOML formatted file."""

    def __init__(self, path: str, valid_keys: Optional[Set[str]] = None) -> None:
        """
        :param path: Relative or absolute path of the config file to use.
        :param valid_keys: When defined, __setitem__ will reject key that are not found in this
                           iterable.
        """
        self._path = os.path.expanduser(path)
        self._valid_keys = valid_keys
        self._toml_dict = {}  # type: Dict[str,Any]

    def exists(self) -> bool:
        """Return True if config exists."""
        return os.path.exists(self._path) and os.path.isfile(self._path)

    def load(self, create: bool = False) -> None:
        """
        Load all data from the config file.
        :param create: When True, the path and config will be created instead of raising an
                       exception.
        """
        try:
            with open(self._path) as f:
                self._toml_dict = toml.load(f)
        except FileNotFoundError:
            if create:
                basedir = os.path.dirname(self._path)
                os.makedirs(basedir)
                open(self._path, 'w').close()
            else:
                raise

    def dump(self) -> None:
        """Dump all values to the config file."""
        with open(self._path, "w+") as f:
            toml.dump(self._toml_dict, f)

    def __getitem__(self, k: str) -> Any:
        return iterutils.get_path(self._toml_dict, k)

    def __setitem__(self, k: str, v: Any) -> None:
        if self._valid_keys and k not in self._valid_keys:
            raise KeyError(f"Key {k} is not allowed")

        nested_tables, k = Conf._extract_table_list(k)
        table = self._toml_dict
        for i in nested_tables:
            table = table.setdefault(i, dict())
        table[k] = v

    def __delitem__(self, k: str) -> None:
        nested_tables, k = Conf._extract_table_list(k)
        table = iterutils.get_path(self._toml_dict, nested_tables)
        del table[k]

    def __iter__(self) -> Iterator:
        raise NotImplemented

    def __len__(self) -> int:
        raise NotImplemented

    @staticmethod
    def _extract_table_list(k: str) -> Tuple[List[str], str]:
        tables = k.split('.')
        return tables, tables.pop(-1)


class ChainConf(collections.ChainMap):
    """A nested collection of dict's that also also supports the Conf() public AP."""

    def exists(self) -> bool:
        """Return result of c.exists() where c is the first Conf instance in the chain."""
        conf = iterutils.first(self.maps, key=lambda x: isinstance(x, Conf))
        return cast(Conf, conf).exists() if conf is not None else False

    def load(self, create: bool = False) -> None:
        """See Conf.load()."""
        for conf in filter(lambda x: isinstance(x, Conf), self.maps):
            cast(Conf, conf).load(create)

    def dump(self) -> None:
        """See Dump.load()."""
        for conf in filter(lambda x: isinstance(x, Conf), self.maps):
            cast(Conf, conf).dump()


# Callable that will always return the global config.
GlobalConf = funcutils.partial(Conf, _GLOBAL_PATH, _GLOBAL_KEYS)  # pylint: disable=invalid-name

# Callable that will always return the default local config.
LocalConf = (  # pylint: disable=invalid-name
    funcutils.partial(ChainConf, Conf(_LOCAL_PATH), GlobalConf()))



"""
Simple ROM Manager - A basic command-line ROM set manager
"""

# These values are exported for case when user runs `pydoc srm` or `dir(srm)`.
# It is generally not a good practice for setup.py to import this module directly.

import pbr.version  # type: ignore

__author__ = 'Patick C. McGinty'
__email__ = 'casey.mcginty@gmail.com'
__version__ = pbr.version.VersionInfo('simple_rom_manager').release_string()



"""
File-like objects used to scan and create directories and archives of files.

The goal of this module is to abstract the tedious logic required read and modify files and
directories within archives, as well as nested archives.
"""

import hashlib
import os
import pathlib
import zlib
from typing import Callable

# Path() -> PosixPath or WindowsPath
# where:
#   PosixPath   -> (Path,PurePosixPath)   -> PurePath -> os.PathLike
#   WindowsPath -> (Path,PureWindowsPath) -> PurePath -> os.PathLike
#
# Path() behaves as a factory in that it delegates to Posix or WindowsPath class initializers.
# However, when used directly PurePath() initializer will delegate to PurePosixPath or
# PureWindowsPath.  Which yields:
#   PurePosixPath   -> PurePath
#   PureWindowsPath -> PurePath
#
# Path uses composition to add a system-aware Accessor as a delegate for all file access functions
# like stat, open, chmod, rmdir, mkdir, rename, etc. Since Python os module already handles the
# internals path differences between windows and posix, pathlib only has one default NormalAccessor.
# Extending to non-path object will require at-least a custom Accessor.
#
# The Accessor is directly injected in Path._init, therefore this method must be patched and/or
# modified by a sub-class. It is probably not sufficient to merely create a new Accessor as the
# abstraction seems to expect all path-like object to behave in the same manor.
#
# Path can be internally initialized with a template accessors, providing a way to read/modify
# path-like objects. This feature allows methods to safely return mutated versions of the instance.
# See resolve() and absolute().
#
# The other key delegation class is Selector, and handles various types of traversals when iterating
# over directories and files. The selector instance is created using pathlib._make_selector()
# internal. It will likely be necessary to extend all of the selector classes so that they can
# detect and initialize custom Path objects.
#
# Selector must be sub-classed and implement _select_from() method to perform the actual Path yield.
# It also depends on Path._make_child_relpath() -> PurePath._from_parsed_parts(). These also will
# need to be made aware of custom path-like objects.

# Step 2:
#   * Create a basic ZipPath object that can be returned by Path()
#   * Update Path.glob() so it can detect and return a ZipPath() object by matching the extension.
#   * Investigate context manager support?


class Path(pathlib.Path):
    """Extend the built-in pathlib.Path class to support additional methods."""

    _READ_SIZE = 1024 * 32

    def __new__(cls, *args: os.PathLike, **kwargs: os.PathLike) -> 'Path':
        if cls is Path:
            cls = WindowsPath if os.name == 'nt' else PosixPath
        return super().__new__(cls, *args, **kwargs)

    def crc(self) -> str:
        """Return CRC hash of file."""
        if self.is_dir():
            raise IsADirectoryError
        crc = 0
        with self.open('rb') as f:
            buf = f.read(self._READ_SIZE)
            while buf:
                crc = zlib.crc32(buf, crc)
                buf = f.read(self._READ_SIZE)
        return '{:08x}'.format(crc & 0xFFFF_FFFF)

    def md5(self) -> str:
        """Return MD5 hash of file."""
        return self._hashlib(hashlib.md5)

    def sha1(self) -> str:
        """Return SHA1 hash of file."""
        return self._hashlib(hashlib.sha1)

    def _hashlib(self, hash_fn: Callable) -> str:
        if self.is_dir():
            raise IsADirectoryError
        hasher = hash_fn()
        with self.open('rb') as f:
            buf = f.read(self._READ_SIZE)
            while buf:
                hasher.update(buf)
                buf = f.read(self._READ_SIZE)
        return str(hasher.hexdigest())


class PosixPath(Path, pathlib.PosixPath):
    """Concrete POSIX Path."""
    __slots__ = ()


class WindowsPath(Path, pathlib.WindowsPath):  # pylint: disable=abstract-method
    """Concrete Windows Path."""
    __slots__ = ()



#!/usr/bin/env python
"""Minimal setuptools entry point configured by Python Build Reasonableness (PBR)."""

from setuptools import setup  # type: ignore

setup(
    setup_requires=['pbr', 'pytest-runner', 'wheel'],
    pbr=True,
    # extra metadata that PBR ignores
    long_description_content_type='text/markdown; charset=UTF-8',
    python_requires='>=3.6',
    tests_require=['pytest'],
)



"""
The default module run when imported from the command line and also the main entry point defined in
setup.py.  Ex:

    python3 -m srm
"""

import click

from . import __version__, cli


@click.group()
@click.version_option(__version__)
def main() -> None:
    """Simple ROM Manager - A basic command-line ROM set manager."""


main.add_command(cli.init)
main.add_command(cli.status)
main(prog_name='srm')  # pylint: disable=unexpected-keyword-arg



"""An auto-increment tool for version strings."""

import unittest.mock
from typing import Any

import click
from click.testing import CliRunner  # type: ignore
from pbr.version import VersionInfo, SemanticVersion  # type: ignore

__version__ = '0.2'


@click.command()
@click.argument('package')
@click.option('--major', 'inc_kwargs', flag_value={'major': True}, help='Increment major number.')
@click.option('--minor', 'inc_kwargs', flag_value={'minor': True}, help='Increment minor number.')
@click.option('--patch', 'inc_kwargs', flag_value={}, default=True, help='Increment patch number.')
def cli(package: str, inc_kwargs: dict) -> None:
    """Bump a MAJOR.MINOR.PATCH version string at the specified index location or 'patch' digit."""

    semver = VersionInfo(package).semantic_version()

    # Pre-decrement a "dev" version to get the real version tuple.
    if 'dev' in semver.version_tuple():
        semver = semver.decrement()

    semver = semver.increment(**inc_kwargs)
    click.echo(semver.release_string(), nl=False)


if __name__ == '__main__':
    cli()  # pylint: disable=no-value-for-parameter


# pylint: disable=missing-docstring,invalid-name
class TestCase(unittest.TestCase):

    def setUp(self):  # type: ignore
        self.runner = CliRunner()

    @unittest.mock.patch.object(VersionInfo, 'semantic_version',
                                return_value=SemanticVersion(0, 0, 0))
    def test_patch_arg(self, _: Any) -> None:
        result = self.runner.invoke(cli, ['dummy', '--patch'])
        self.assertEqual('0.0.1', result.output)

    @unittest.mock.patch.object(VersionInfo, 'semantic_version',
                                return_value=SemanticVersion(0, 0, 0))
    def test_minor_arg(self, _: Any) -> None:
        result = self.runner.invoke(cli, ['dummy', '--minor'])
        self.assertEqual('0.1.0', result.output)

    @unittest.mock.patch.object(VersionInfo, 'semantic_version',
                                return_value=SemanticVersion(0, 0, 0))
    def test_major_arg(self, _: Any) -> None:
        result = self.runner.invoke(cli, ['dummy', '--major'])
        self.assertEqual('1.0.0', result.output)

    @unittest.mock.patch.object(VersionInfo, 'semantic_version',
                                return_value=SemanticVersion(0, 0, 1, None, None, 1))
    def test_patch_arg_on_dev_ver(self, _: Any) -> None:
        result = self.runner.invoke(cli, ['dummy', '--patch'])
        self.assertEqual('0.0.1', result.output)



"""
Test cases for DAT XML import and processing.
"""

from itertools import zip_longest
from unittest.mock import ANY, mock_open, patch, sentinel

import pytest

from srm.dat import ROM, DatafileXml, Game

PATH = sentinel.PATH

DAT_01 = dict(
    id='final-burn-simple',
    xml="""\
<?xml version="1.0"?>
<!DOCTYPE datafile PUBLIC "-//FB Alpha//DTD ROM Management Datafile//EN" "http://www.logiqx.com/Dats/datafile.dtd">
<datafile>
    <header>
        <name>FB Alpha - SuprGrafx Games</name>
        <description>FB Alpha v0.2.97.42 SuprGrafx Games</description>
        <category>Standard DatFile</category>
        <version>0.2.97.42</version>
        <author>FB Alpha</author>
        <homepage>https://www.fbalpha.com/</homepage>
        <url>https://www.fbalpha.com/</url>
        <clrmamepro forcenodump="ignore"/>
    </header>
    <game name="1941">
        <description>1941 - Counter Attack</description>
        <year>1991</year>
        <manufacturer>Hudson</manufacturer>
        <rom name="1941 - counter attack (japan).pce" size="1048576" crc="8c4588e2"/>
    </game>
    <game name="aldynes">
        <description>Aldynes</description>
        <year>1991</year>
        <manufacturer>Hudson</manufacturer>
        <rom name="aldynes (japan).pce" size="1048576" crc="4c2126b0"/>
    </game>
</datafile>
""",
    header=dict(
        name='FB Alpha - SuprGrafx Games',
        description='FB Alpha v0.2.97.42 SuprGrafx Games',
        version='0.2.97.42',
        author='FB Alpha',
        homepage='https://www.fbalpha.com/',
        url='https://www.fbalpha.com/',
    ),
    games=[
        Game(
            name="1941",
            description="1941 - Counter Attack",
            year="1991",
            manufacturer="Hudson",
            roms=ANY,
        ),
        Game(
            name="aldynes",
            description="Aldynes",
            year="1991",
            manufacturer="Hudson",
            roms=ANY,
        ),
    ],
    roms=[
        ROM(name="1941 - counter attack (japan).pce", size="1048576", crc="8c4588e2"),
        ROM(name="aldynes (japan).pce", size="1048576", crc="4c2126b0"),
    ],
)

DAT_02 = dict(
    id='final-burn-clone',
    xml="""\
<?xml version="1.0"?>
<!DOCTYPE datafile PUBLIC "-//FB Alpha//DTD ROM Management Datafile//EN" "http://www.logiqx.com/Dats/datafile.dtd">
<datafile>
    <header>
        <name>FB Alpha - Arcade Games</name>
        <description>FB Alpha v0.2.97.42 Arcade Games</description>
        <category>Standard DatFile</category>
        <version>0.2.97.42</version>
        <author>FB Alpha</author>
        <homepage>https://www.fbalpha.com/</homepage>
        <url>https://www.fbalpha.com/</url>
        <clrmamepro forcenodump="ignore"/>
    </header>
    <game name="gtmrb" cloneof="gtmr" romof="gtmr">
        <description>1000 Miglia: Great 1000 Miles Rally (94/05/26)</description>
        <year>1994</year>
        <manufacturer>Kaneko</manufacturer>
        <rom name="mmp0x1.u514" size="524288" crc="6c163f12"/>
        <rom name="mmp1x1.u513" size="524288" crc="424dc7e1"/>
        <rom name="mmd0x1.u124" size="131072" crc="3d7cb329"/>
        <rom name="mm-200-402-s0.bin" merge="mm-200-402-s0.bin" size="2097152" crc="c0ab3efc"/>
        <rom name="mm-201-403-s1.bin" merge="mm-201-403-s1.bin" size="2097152" crc="cf6b23dc"/>
        <rom name="mm-202-404-s2.bin" merge="mm-202-404-s2.bin" size="2097152" crc="8f27f5d3"/>
        <rom name="mm-203-405-s3.bin" merge="mm-203-405-s3.bin" size="524288" crc="e9747c8c"/>
        <rom name="mms1x1.u30" size="131072" crc="9463825c"/>
        <rom name="mms0x1.u29" size="131072" crc="bd22b7d2"/>
        <rom name="mm-300-406-a0.bin" merge="mm-300-406-a0.bin" size="2097152" crc="b15f6b7f"/>
        <rom name="mm-100-401-e0.bin" merge="mm-100-401-e0.bin" size="1048576" crc="b9cbfbee"/>
    </game>
    <game name="gtmr">
        <description>1000 Miglia: Great 1000 Miles Rally (94/07/18)</description>
        <year>1994</year>
        <manufacturer>Kaneko</manufacturer>
        <rom name="u2.bin" size="524288" crc="031799f7"/>
        <rom name="u1.bin" size="524288" crc="6238790a"/>
        <rom name="mmd0x2.u124.bin" size="131072" crc="3d7cb329"/>
        <rom name="mm-200-402-s0.bin" size="2097152" crc="c0ab3efc"/>
        <rom name="mm-201-403-s1.bin" size="2097152" crc="cf6b23dc"/>
        <rom name="mm-202-404-s2.bin" size="2097152" crc="8f27f5d3"/>
        <rom name="mm-203-405-s3.bin" size="524288" crc="e9747c8c"/>
        <rom name="mms1x2.u30.bin" size="131072" crc="b42b426f"/>
        <rom name="mms0x2.u29.bin" size="131072" crc="bd22b7d2"/>
        <rom name="mm-300-406-a0.bin" size="2097152" crc="b15f6b7f"/>
        <rom name="mm-100-401-e0.bin" size="1048576" crc="b9cbfbee"/>
    </game>
</datafile>
""",
    header=dict(
        name='FB Alpha - Arcade Games',
        description='FB Alpha v0.2.97.42 Arcade Games',
        version='0.2.97.42',
        author='FB Alpha',
        homepage='https://www.fbalpha.com/',
        url='https://www.fbalpha.com/',
    ),
    games=[
        Game(
            name="gtmrb",
            description="1000 Miglia: Great 1000 Miles Rally (94/05/26)",
            year="1994",
            manufacturer="Kaneko",
            cloneof="gtmr",
            romof="gtmr",
            roms=ANY,
        ),
        Game(
            name="gtmr",
            description="1000 Miglia: Great 1000 Miles Rally (94/07/18)",
            year="1994",
            manufacturer="Kaneko",
            cloneof="",
            romof="",
            roms=ANY,
        ),
    ],
    roms=[
        ROM(name="mm-100-401-e0.bin", merge="mm-100-401-e0.bin", size="1048576", crc="b9cbfbee"),
        ROM(name="mm-100-401-e0.bin", size="1048576", crc="b9cbfbee"),
        ROM(name="mm-200-402-s0.bin", merge="mm-200-402-s0.bin", size="2097152", crc="c0ab3efc"),
        ROM(name="mm-200-402-s0.bin", size="2097152", crc="c0ab3efc"),
        ROM(name="mm-201-403-s1.bin", merge="mm-201-403-s1.bin", size="2097152", crc="cf6b23dc"),
        ROM(name="mm-201-403-s1.bin", size="2097152", crc="cf6b23dc"),
        ROM(name="mm-202-404-s2.bin", merge="mm-202-404-s2.bin", size="2097152", crc="8f27f5d3"),
        ROM(name="mm-202-404-s2.bin", size="2097152", crc="8f27f5d3"),
        ROM(name="mm-203-405-s3.bin", merge="mm-203-405-s3.bin", size="524288", crc="e9747c8c"),
        ROM(name="mm-203-405-s3.bin", size="524288", crc="e9747c8c"),
        ROM(name="mm-300-406-a0.bin", merge="mm-300-406-a0.bin", size="2097152", crc="b15f6b7f"),
        ROM(name="mm-300-406-a0.bin", size="2097152", crc="b15f6b7f"),
        ROM(name="mmd0x1.u124", size="131072", crc="3d7cb329"),
        ROM(name="mmd0x2.u124.bin", size="131072", crc="3d7cb329"),
        ROM(name="mmp0x1.u514", size="524288", crc="6c163f12"),
        ROM(name="mmp1x1.u513", size="524288", crc="424dc7e1"),
        ROM(name="mms0x1.u29", size="131072", crc="bd22b7d2"),
        ROM(name="mms0x2.u29.bin", size="131072", crc="bd22b7d2"),
        ROM(name="mms1x1.u30", size="131072", crc="9463825c"),
        ROM(name="mms1x2.u30.bin", size="131072", crc="b42b426f"),
        ROM(name="u1.bin", size="524288", crc="6238790a"),
        ROM(name="u2.bin", size="524288", crc="031799f7"),
    ],
)

DAT_03 = dict(
    id='no-intro',
    xml="""\
<?xml version="1.0"?>
<!DOCTYPE datafile PUBLIC "-//Logiqx//DTD ROM Management Datafile//EN" "http://www.logiqx.com/Dats/datafile.dtd">
<datafile>
    <header>
        <name>Nintendo - Game Boy</name>
        <description>Nintendo - Game Boy</description>
        <version>20171226-085946</version>
        <author>C. V. Reynolds, Densetsu, xuom2</author>
        <homepage>No-Intro</homepage>
        <url>http://www.no-intro.org</url>
    </header>
    <game name="Battletoads (Japan)">
        <description>Battletoads (Japan)</description>
        <rom name="Battletoads (Japan).gb" size="131072" crc="331CF7DE" md5="3D57E0391C8191C105A4F015A0C103E9" sha1="666ED5D34F508C8805A67F4400FC01A1F2817E03"/>
    </game>
</datafile>
""",
    header=dict(
        name='Nintendo - Game Boy',
        description='Nintendo - Game Boy',
        version='20171226-085946',
        author='C. V. Reynolds, Densetsu, xuom2',
        homepage='No-Intro',
        url='http://www.no-intro.org',
    ),
    games=[
        Game(
            name="Battletoads (Japan)",
            description="Battletoads (Japan)",
            roms=ANY,
        ),
    ],
    roms=[
        ROM(name="Battletoads (Japan).gb",
            size="131072",
            crc="331CF7DE",
            md5="3D57E0391C8191C105A4F015A0C103E9",
            sha1="666ED5D34F508C8805A67F4400FC01A1F2817E03")
    ],
)


@pytest.mark.parametrize("xml,header", [
    pytest.param(DAT_01['xml'], DAT_01['header'], id=DAT_01['id']),
    pytest.param(DAT_02['xml'], DAT_02['header'], id=DAT_02['id']),
    pytest.param(DAT_03['xml'], DAT_03['header'], id=DAT_03['id']),
])
def test_load_header_from_dat(xml, header):
    with patch('xml.etree.ElementTree.open', mock_open(read_data=xml)) as mock_file:
        dat = DatafileXml(PATH)

    mock_file.assert_called_with(PATH, ANY)
    assert dat.name == header['name']
    assert dat.description == header['description']
    assert dat.version == header['version']
    assert dat.author == header['author']
    assert dat.homepage == header['homepage']
    assert dat.url == header['url']


@pytest.mark.parametrize("xml,exp_games", [
    pytest.param(DAT_01['xml'], DAT_01['games'], id=DAT_01['id']),
    pytest.param(DAT_02['xml'], DAT_02['games'], id=DAT_02['id']),
    pytest.param(DAT_03['xml'], DAT_03['games'], id=DAT_03['id']),
])
def test_load_games_from_dat(xml, exp_games):
    found_games = set()
    with patch('xml.etree.ElementTree.open', mock_open(read_data=xml)) as mock_file:
        dat = DatafileXml(PATH)
    for game in dat.games:
        assert game in exp_games
        assert game not in found_games
        found_games.add(game)


@pytest.mark.parametrize("xml,exp_roms", [
    pytest.param(DAT_01['xml'], DAT_01['roms'], id=DAT_01['id']),
    pytest.param(DAT_02['xml'], DAT_02['roms'], id=DAT_02['id']),
    pytest.param(DAT_03['xml'], DAT_03['roms'], id=DAT_03['id']),
])
def test_load_roms_from_dat(xml, exp_roms):
    found_roms = set()
    with patch('xml.etree.ElementTree.open', mock_open(read_data=xml)) as mock_file:
        dat = DatafileXml(PATH)
        for game in dat.games:
            for rom in game.roms:
                assert rom in exp_roms
                assert rom not in found_roms
                found_roms.add(rom)



"""
Load and query XML data that describes ROM game collections.
"""

from typing import Any, Iterable, Set
from xml.etree import ElementTree

import attr  # type: ignore

# import boltons.cacheutils  # type: ignore


class DatafileXml:
    """Parse a DAT XML file (datafile DTD)."""

    def __init__(self, file_path: str) -> None:
        tree = ElementTree.parse(file_path,)
        self._root = tree.getroot()
        header = self._root.find('header')
        if header is None:
            raise ElementTree.ParseError('DAT file does not contain a valid "header"')
        self._header = Header.from_xml(header)

    @property
    def name(self) -> str:
        """DAT <header><name> field."""
        return str(self._header.name)

    @property
    def description(self) -> str:
        """DAT <header><description> field."""
        return str(self._header.description)

    @property
    def version(self) -> str:
        """DAT <header><version> field."""
        return str(self._header.version)

    @property
    def author(self) -> str:
        """DAT <header><author> field."""
        return str(self._header.author)

    @property
    def date(self) -> str:
        """DAT <header><date> field."""
        return str(self._header.date)

    @property
    def homepage(self) -> str:
        """DAT <header><homepage> field."""
        return str(self._header.homepage)

    @property
    def url(self) -> str:
        """DAT <header><url> field."""
        return str(self._header.url)

    @property
    def games(self) -> Iterable['Game']:
        """Iterable of <game> tag elements from DAT."""
        for game in self._root.iter('game'):
            roms = frozenset({ROM.from_xml(r) for r in game.iter('rom')})
            yield Game.from_xml(game, roms=roms)

    # @boltons.cacheutils.cachedproperty
    # def crcs(self) -> Set[str]:
    #     return set(r.crc for r in self.roms)
    #
    # @boltons.cacheutils.cachedproperty
    # def md5s(self) -> Set[str]:
    #     return set(r.md5 for r in self.roms)

    # def matching(self, crc=None, md5=None):
    #     if crc:
    #         return self._crc_index[crc]
    #     elif md5:
    #         return self._md5_index[md5]
    #     else:
    #         raise ValueError('Method called with no arguments')


class XmlToAttrs:  # pylint: disable=too-few-public-methods
    """Helper to generically map XML fields from an element and create an attrs class."""

    @classmethod
    def from_xml(cls, root: ElementTree.Element, **cls_kwargs: Any) -> Any:
        """
        Dynamically extract attrs fields from XML root element attributes and sub-tags to initialize
        a new cls() instance.

        :param root: XML element to extract class fields from.
        :param cls_kwargs: Additional kwargs passed into the class constructor.
        """
        # Create a combined dict of root attributes and sub-tags.
        cls_kwargs.update(root.attrib)
        cls_kwargs.update({i.tag: i.text or '' for i in root.iter() if i is not root})
        # Init the class using only the expected fields defined by attrs declarations.
        init_kwds = list(i.name for i in iter(attr.fields(cls)))
        return cls(**{k: v for k, v in cls_kwargs.items() if k in init_kwds})  # type: ignore


@attr.s(frozen=True, slots=True, auto_attribs=True)  # pylint: disable=too-few-public-methods
class Header(XmlToAttrs):
    """ROM Datfile XML header info."""
    name: str  # official name of DAT
    description: str  # extended name
    version: str  # numeric version (e.g. 2.40.900  or 20171210-4500)
    author: str  # contributors or release group
    date: str = ''
    homepage: str = ''
    url: str = ''


@attr.s(frozen=True, slots=True, auto_attribs=True)  # pylint: disable=too-few-public-methods
class ROM(XmlToAttrs):
    """ROM object defined for one or more Games."""
    name: str  # file name
    size: int  # bytes
    merge: str = ''  # merge file name, indicates file was merged from "cloneof"
    crc: str = ''
    md5: str = ''
    sha1: str = ''


@attr.s(frozen=True, slots=True, auto_attribs=True)  # pylint: disable=too-few-public-methods
class Game(XmlToAttrs):
    """Game record with one or more ROMs."""
    name: str  # file name
    description: str  # common name (can be same as 'name')
    roms: Set[ROM]  # collection of ROMs required by the game

    cloneof: str = ''
    isbios: bool = False
    manufacturer: str = ''
    romof: str = ''
    year: str = ''
